/*
 * JNU - Java Native Utility stuff.
 *
 * The +name+ JNU is from Liang, The Java Native Interface, Addison Wesley 1999,
 * where it refers to a family of 'C' functions.
 * The C++ classes herein are +my own+, in both design and implementation.
 *
 * Note that none of these classes stores a JNIEnv pointer,
 * which means that they are all thread-safe.
 * Instead the JNIEnv is always acquired from the JVM on demand.
 *
 * They are all primarily designed to be used on the stack,
 * rather than via new and delete, which is strongly discouraged
 *
 * This work is not to be confused with the article and source code by
 * Gabrilovich and Finkelstein (C/C++ Users Journal, 19(1):10-21, January 2001).
 * Compared with that, this work:
 *
 *	- is somewhat simpler ;-).
 *	- supports a Java-like style of programming, e.g. by providing Java APIs
 *		where possible, rather than providing a C++-like style with
 *		widespread use of constructors and overloaded = operators.
 *	- is considerably more free of bugs ;-)).
 *	- avoids the resource-copying issue altogether by simply prohibiting
 *		the copying of classes which allocate resources behind the scenes.
 *	- relies exclusively on C++ classes and templates,
 *		and makes no use of the preprocessor whatsoever.
 *	- provides considerably more coverage of the JNI API.
 *
 * Typical usage: a method that just calls self.setValue(value):
 *
 * JNIEXPORT void JNICALL Java_com_mycorp_mypackage_myclass_mynative0(JNIEnv *env, jobject self, jint value)
 * {
 *	JNULocalFrame	frame;
 *	JNUObject<jobject> jnuSelf = self;
 *  JNUClass  clazz = jnuSelf.getClass();
 *  JNUVoidMethod method = clazz.getMethod("setValue","(I)V");
 *	method.invoke(jnuSelf, value);
 * }
 *
 * You get the idea. It is very much like Java reflection.
 *
 * Copyright (c) Esmond Pitt, 2008. All rights reserved.
 */
#include "jnu.hpp"

#ifdef _MSC_VER
#pragma warning(push, 1)
//#pragma warning(disable : 4786)
#endif

/*****************************************************************************
* JNI_OnLoad - called when this library is loaded.
JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *jvm, void *reserved)
{
	printf("JNI_OnLoad called\n"); fflush(stdout);
	JNUJVM::jvm = jvm;
	return JNI_VERSION_1_2;
}
*/

/*****************************************************************************
* JNI_OnUnload - called when this library is unloaded.
JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *jvm, void *reserved)
{
	printf("JNI_OnUnload called\n"); fflush(stdout);
	JNUJVM::jvm = 0;
}
*/

// JNUJVM
JavaVM *JNUJVM::jvm;

JNUEnv JNUJVM::env()
{
	if (jvm == 0)
		getCreatedJavaVMs(jvm, 1);
	void	*env;
	jvm->GetEnv(&env, JNI_VERSION_1_4);
	return (JNIEnv*)env;
}

jint	JNUJVM::getCreatedJavaVMs(JavaVM *&jvms, jint size) throw(java::jnu::PendingJavaException)
{
	jint	count;
	jint	rc = JNI_GetCreatedJavaVMs(&jvms, size, &count);
	return count;
}

jint	JNUJVM::getVersion()
{
	return env()->GetVersion();
}

// JNUEnv
JNUEnv::	JNUEnv(JNIEnv *env)
	: _env(env)
{
}

JNUEnv::~JNUEnv()
{
}

jclass JNUEnv::getWeakGlobalRef(const JNUObject<jobject> &object, jweak &weak)
{
	jclass	cls = (jclass)weak;
	if (weak == 0 || _env->IsSameObject(weak, 0))
	{
		// uninitialized or has been GC'd
		cls = _env->GetObjectClass(object);
		// No need for exception check here as all objects have classes.
		weak = _env->NewWeakGlobalRef(cls);
	}
	return cls;
}

jclass JNUEnv::getWeakGlobalRef(const char *className, jweak &weak)
{
	jclass	cls = (jclass)weak;
	if (weak == 0 || _env->IsSameObject(weak, 0))
	{
		// uninitialized or has been GC'd
		cls = _env->FindClass(className);
		// Exception check here, as not all names name real classes.
		if (cls == 0 /*=>env->ExceptionCheck()*/)
			return 0;
		weak = _env->NewWeakGlobalRef(cls);
	}
	return cls;
}

void JNUEnv::throwException(const char *excClassName, const char *message)
{
	// Declare a local frame
//	JNULocalFrame	frame();
	// printf("throwing exception %s(%s)\n", excClassName, message); fflush(stdout);
	// create an instance of the specified exception
	JNUClass exceptionClass = JNUClass::forName(excClassName);
	
	if(exceptionClass != NULL) 
		// throw the new exception back to the java wrapper
		_env->ThrowNew(exceptionClass, message);
	
	// free the local reference. if exception is still null, delete is a noop.
	_env->DeleteLocalRef(exceptionClass);
	
}

void	JNUEnv::checkPendingException() const
	throw(PendingJavaException)
{
	if (_env->ExceptionCheck())
	{
		_env->ExceptionDescribe();
		throw PendingJavaException();
	}
}

// JNULocalFrame
JNULocalFrame::JNULocalFrame(jint capacity) throw (java::jnu::PendingJavaException)
	: JNUEnv(JNUJVM::env()),
	_popped(0)
{
	// printf("pushing local frame pre exception check\n"); fflush(stdout);
	_env->PushLocalFrame(capacity);
	checkPendingException();
	// printf("pushing local frame post exception check\n"); fflush(stdout);
}

JNULocalFrame::~JNULocalFrame()
{
	if (!_popped)
	{
		// printf("popping a local frame\n"); fflush(stdout);
		_env->PopLocalFrame(0);
		// printf("popped a local frame\n"); fflush(stdout);
		// Checking for an exception after PopLocalFrame()
		// is fairly pointless, as the semantics of C++
		// constructors and destructors guarantee the right kind of symmetry
		// with PushLocalFrame().
		// Also it would mean that any exception arising in any
		// code block that includes a JNULocalFrame would be reported
		// here, which is not the intent.
//		if (_env->ExceptionCheck())
//		{
//			fprintf(stderr, "JNULocalFrame::~JNULocalFrame(): exception detected in JNI code\n");
			// WWWWWWWWWW
			// WARNING
			// ExceptionDescribe() clears the exception!
			// THIS IS NOT DOCUMENTED ANYWHERE!!!!!!!!!
//			_env->ExceptionDescribe();
//			if (_env->ExceptionCheck())
//				fprintf(stderr, "exception is still pending\n");
//			else
//				fprintf(stderr, "**** Exception is NOT still pending!\n");
//			fflush(stdout);
//			// Don't throw exceptions from destructors!
//			// throw PendingJavaException();
//		}
	}
}

void	JNULocalFrame::ensureLocalCapacity(jint capacity) throw (java::jnu::PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	env->EnsureLocalCapacity(capacity);
	env.checkPendingException();
}

// synchronized
synchronized::synchronized(const JNUObject<jobject> &object)
	: _object(object)
{
	JNUJVM::env()->MonitorEnter(object);
}

synchronized::~synchronized()
{
	JNUJVM::env()->MonitorExit(_object);
}

// JNUClass
JNUClass::JNUClass(jclass clazz) throw (java::jnu::PendingJavaException)
	: JNUObject<jclass>(clazz)
{
//	printf("JNUClass::JNUClass: name=%s\n", (const char*)getName()); fflush(stdout);
}

JNUClass	JNUClass::forName(const char *className) throw (java::jnu::PendingJavaException)
{
//	printf("JNUClass::forName(%s)\n", className); fflush(stdout);
	JNUClass	cls(JNUJVM::env()->FindClass(className));
	JNUJVM::env().checkPendingException();
	return cls;
}

JNUClass	JNUClass::forName(const char *className, jweak &cls) throw (java::jnu::PendingJavaException)
{
	return JNUClass(JNUWeakGlobalRef<jclass>(className, cls));
}

JNUConstructor	JNUClass::getConstructor(const char *signature) const throw (java::jnu::PendingJavaException)
{
	JNUConstructor ctor(getMethod("<init>", signature));
	JNUJVM::env().checkPendingException();
	return ctor;
}

JNUFieldID	JNUClass::getField(const char *name, const char *signature) const throw (java::jnu::PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	// Use *this to cause operator jclass() to be called.
	jfieldID	fid = env->GetFieldID(*this, name, signature);
	if (fid == 0)
	{
		// printf("NoSuchField %s%s\n", name, signature); fflush(stdout);
		env->ExceptionDescribe();
		throw PendingJavaException();
	}
	return JNUFieldID(name, fid, *this);
}

JNUFieldID JNUClass::getStaticField(const char *name, const char *signature) const throw (java::jnu::PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	// Use *this to cause operator jclass() to be called.
	jfieldID	fid = env->GetStaticFieldID(*this, name, signature);
	if (fid == 0)
	{
		// printf("NoSuchField %s%s\n", name, signature); fflush(stdout);
		env->ExceptionDescribe();
		throw PendingJavaException();
	}
	return JNUFieldID(name, fid, *this, JNI_TRUE);
}

JNUMethodID	JNUClass::getMethod(const char *name, const char *signature) const throw (java::jnu::PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	// Use *this to cause operator jclass() to be called.
//	printf("getting methodID %x %s%s\n", _object, name, signature); fflush(stdout);
	jmethodID	mid = env->GetMethodID(*this, name, signature);
	if (mid == 0)
	{
		// printf("NoSuchMethod %s%s\n", name, signature); fflush(stdout);
		env->ExceptionDescribe();
		throw PendingJavaException();
	}
	// printf("methodID(%s%s=%8.8x\n", name, signature, mid);
	return JNUMethodID(name, mid, *this);
}

JNUMethodID	JNUClass::getStaticMethod(const char *name, const char *signature) const throw (java::jnu::PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	// Use *this to cause operator jclass() to be called.
	jmethodID	mid = env->GetStaticMethodID(*this, name, signature);
	if (mid == 0)
	{
		// printf("NoSuchMethod %s%s\n", name, signature); fflush(stdout);
		env->ExceptionDescribe();
		throw PendingJavaException();
	}
	// printf("static methodID(%s%s=%8.8x\n", name, signature, mid);
	return JNUMethodID(name, mid, *this, JNI_TRUE);
}

#if NOTYET
JNUString	JNUClass::getName() const
{
	return (jstring)JNUObjectMethod(getMethod("getName", "()Ljava/lang/String;")).invoke((jobject)_object);
}
#endif

jclass	JNUClass::getSuperclass() const
{
	return JNUJVM::env()->GetSuperclass(*this);
}

jboolean	JNUClass::isInstance(const jobject object) const
{
	return JNUJVM::env()->IsInstanceOf(object, *this);
}

jboolean	JNUClass::isAssignableFrom(const JNUClass &clazz) const
{
	return JNUJVM::env()->IsAssignableFrom(clazz, *this);
}

jobject	JNUClass::newInstance() const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	// Using the default constructor.
	jobject	object = JNUObjectMethod(getConstructor()).invoke(*this);
	env.checkPendingException();
	return object;
}

// JNUThrowable
JNUThrowable::JNUThrowable(jthrowable throwable)
	: JNUObject<jthrowable>(throwable)
{
}

JNUThrowable::JNUThrowable(const char *className, const char *message)
	: JNUObject<jthrowable>(0)
{
	JNUJVM::env()->ThrowNew(JNUClass::forName(className), message);
}

void JNUThrowable::throwIt()
{
	JNUJVM::env()->Throw(*this);
}

void	JNUThrowable::printStackTrace()
{
	JNUJVM::env()->ExceptionDescribe();
}

jboolean	JNUThrowable::isExceptionPending()
{
	return JNUJVM::env()->ExceptionCheck();
}

void	JNUThrowable::clear()
{
	JNUJVM::env()->ExceptionClear();
}

jthrowable	JNUThrowable::currentException()
{
	return JNUJVM::env()->ExceptionOccurred();
}

// JNUString
JNUString::	JNUString(jstring string)
	: super(string),
	_unicode(0),
	utf(0)
{
}

JNUString::JNUString(const jchar *unicode, jsize length)
	: super(JNUJVM::env()->NewString(unicode, length)),
	_unicode(0),
	utf(0)
{
	JNUJVM::env().checkPendingException();
}

JNUString::JNUString(const char *utf)
	: JNUObject<jstring>(JNUJVM::env()->NewStringUTF(utf)),
	_unicode(0),
	utf(0)
{
	JNUJVM::env().checkPendingException();
}

JNUString::~JNUString()
	throw()
{
	JNUEnv	env = JNUJVM::env();
	if (_unicode)	// have been acquired
	{
		//			printf("ReleaseStringChars(%8.8x)\n", _bytes); fflush(stdout);
		env->ReleaseStringChars(_object, _unicode);
		if (env->ExceptionCheck())
		{
			env->ExceptionDescribe();
			// Don't throw exceptions from destructors!
			// throw PendingJavaException();
		}
	}
	if (utf)	// have been acquired
	{
		env->ReleaseStringUTFChars(_object, utf);
		if (env->ExceptionCheck())
		{
			env->ExceptionDescribe();
			// Dont' throw exceptions from destructors!
			// throw PendingJavaException();
		}
	}
}

jsize	JNUString::getLength() const
{
	return _object ? JNUJVM::env()->GetStringLength(_object) : 0;
}

const jchar *JNUString::getChars()
{
	// NB cope with the null string
	if (_object != 0 && _unicode == 0)
	{
		_unicode = JNUJVM::env()->GetStringChars(_object, 0);
	}
	return _unicode;
}

void JNUString::getRegion(jsize start, jsize len, jchar *buf)
{
	JNUJVM::env()->GetStringRegion(_object, start, len, buf);
}

const char *JNUString::getUTFChars()
{
	// NB cope with the null string
	if (_object != 0 && utf == 0)
	{
		utf = JNUJVM::env()->GetStringUTFChars(_object, 0);
	}
	return utf;
}

jsize	JNUString::getLengthUTF() const
{
	return _object ? JNUJVM::env()->GetStringUTFLength(_object) : 0;
}

void JNUString::getRegion(jsize start, jsize len, char *buf)
{
	JNUJVM::env()->GetStringUTFRegion(_object, start, len, buf);
}

// JNURef
JNURef::JNURef(jobject ref)
: _ref(ref)
{
}

JNURef::~JNURef()
{
}

// JNUGlobalRef
JNUGlobalRef::JNUGlobalRef(jobject ref)
: JNURef(JNUJVM::env()->NewGlobalRef(_ref))
{
}

JNUGlobalRef::~JNUGlobalRef()
{
	JNUJVM::env()->DeleteGlobalRef(_ref);
}

JNULocalRef	JNUGlobalRef::getLocalRef() const
{
	return JNULocalRef(*this);
}

// JNULocalRef
JNULocalRef::JNULocalRef(jobject ref)
: JNURef(JNUJVM::env()->NewLocalRef(_ref))
{
}

JNULocalRef::~JNULocalRef()
{
	JNUJVM::env()->DeleteLocalRef(_ref);
}

// JNUWeakGlobalRef
template<class JType>
JNUWeakGlobalRef<JType>::JNUWeakGlobalRef(const char *className, jweak &weak)
: JNULocalRef(0),
	_className(className),
	_weak(weak)
{
}

// Copy constructor
template<class JType>
JNUWeakGlobalRef<JType>::JNUWeakGlobalRef(const JNUWeakGlobalRef<JType> &that)
	: JNULocalRef(that),
	_className(that._className),
	_weak(that._weakRef)
{
	// MOST IMPORTANT: don't copy that._localRef, make my own when needed.
	_ref = 0;
}

template<class JType>
JNUWeakGlobalRef<JType>::~JNUWeakGlobalRef()
{
	JNUJVM::env()->DeleteWeakGlobalRef((jweak)_ref);
}

template<> JNUWeakGlobalRef<jclass>::operator jclass() const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	if (_weak == 0 || env->IsSameObject(_weak, 0))
	{
		// uninitialized or has been GC'd
		jclass cls = JNUClass::forName(_className);
		// Now convert the static reference to a weak global reference.
		const_cast<JNUWeakGlobalRef<jclass>*>(this)->_weak = env->NewWeakGlobalRef(cls);
		// Acquire a LocalRef for use in this scope
		const_cast<JNUWeakGlobalRef<jclass>*>(this)->_ref = (jclass)env->NewLocalRef(cls);
	}
	return (jclass)_ref;
}

//JNUAccessibleObject method table
template<> jobject (JNIEnv::*const JNUAccessibleObject<jfieldID>::ToReflected)(jclass clazz, jfieldID fieldID, jboolean isStatic)		= &JNIEnv::ToReflectedField;
template<> jobject (JNIEnv::*const JNUAccessibleObject<jmethodID>::ToReflected)(jclass clazz, jmethodID methodID, jboolean isStatic)	= &JNIEnv::ToReflectedMethod;

// JNUConstructor
JNUConstructor::JNUConstructor(const JNUMethodID &methodID)
	: JNUAccessibleObject<jmethodID>(methodID),
	JNUMember<jmethodID>(methodID)
{
}

jobject	JNUConstructor::newInstance(jclass cls, ...)
	throw (java::jnu::PendingJavaException)
{
	// printf("newInstance(cid=%8.8x, ...)\n", _methodID); fflush(stdout);
	JNUEnv	env = JNUJVM::env();
	va_list	args;
	va_start(args, cls);
	jobject	object = env->NewObjectV(cls, id(), args);
	env.checkPendingException();
	return object;
}

jobject	JNUConstructor::newInstance(const JNUClass &cls, const jvalue *args)
	throw (java::jnu::PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	// printf("newInstance(cid=%8.8x, jvalue *\n", _methodID); fflush(stdout);
	jobject	object = env->NewObjectA(cls, id(), args);
	env.checkPendingException();
	return object;
}

jobject	JNUConstructor::newInstance(const JNUClass &cls, va_list args)
	throw (java::jnu::PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	// printf("newInstance(cid=%8.8x, va_list)\n", _methodID); fflush(stdout);
	jobject	object = env->NewObjectV(cls, id(), args);
	env.checkPendingException();
	return object;
}

#if NOTYET
// JNUPrimitiveArrayCritical
JNUPrimitiveArrayCritical::JNUPrimitiveArrayCritical(JArrayType array, int mode = 0)
	throw(PendingJavaException)
	: JNUPrimitiveArray<JArrayType, NativeType>(array, mode)
{
//		JNUJVM::env().checkPendingException();
	if (mode != 0 && mode == JNI_ABORT)
		throw JNUThrowable("java/lang/IllegalArgumentException", "JNUPrimitiveArrayCritical.<init>: mode");
	_elements = (NativeType*)JNUJVM::env()->GetPrimitiveArrayCritical(array, &_isCopy);
}

JNUPrimitiveArrayCritical::~JNUPrimitiveArrayCritical()
{
	JNUJVM::env()->ReleasePrimitiveArrayCritical(*this, _elements, _mode);
	// Stop the base class from doing anything else with _elements.
	_elements = 0;
}

void JNUPrimitiveArrayCritical::commit()
	throw(PendingJavaException)
{
	throw JNUUnsupportedOperationException("JNUPrimitiveArrayCritical::commit()");
}

#endif

// JNUObjectArray
JNUObjectArray::JNUObjectArray(jobjectArray array)
	throw(PendingJavaException)
	: JNUArray<jobjectArray>(array)
{
	// printf("JNUObjectArray::JNUObjectArray() pre exception check\n"); fflush(stdout);
	JNUJVM::env().checkPendingException();
	// printf("JNUObjectArray::JNUObjectArray() post exception check\n"); fflush(stdout);
}

JNUObjectArray::JNUObjectArray(jsize length, const JNUClass &cls, jobject initial)
	throw(PendingJavaException)
	: JNUArray<jobjectArray>(JNUJVM::env()->NewObjectArray(length, cls, initial))
{
	// printf("JNUObjectArray::JNUObjectArray() pre exception check\n"); fflush(stdout);
	JNUJVM::env().checkPendingException();
	// printf("JNUObjectArray::JNUObjectArray() post exception check\n"); fflush(stdout);
}

JNUObjectArray::~JNUObjectArray() throw()
{
}


jobject	JNUObjectArray::get(jsize n) const
	throw (java::jnu::PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	jobject value = env->GetObjectArrayElement(_object, n);
	env.checkPendingException();
	return value;
}

const jobject	JNUObjectArray::operator[](int n) const
	throw (java::jnu::PendingJavaException)
{
	return get(n);
}

void	JNUObjectArray::set(jsize n, jobject value)
	throw (java::jnu::PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	env->SetObjectArrayElement(_object, n, value);
	env.checkPendingException();
}

// JNUBuffer
JNUBuffer::JNUBuffer(const JNUObject<jobject> &object)
	: JNUObject<jobject>(object)
{
}

jsize	JNUBuffer::getCapacity() const
	throw (java::jnu::PendingJavaException)
{
	return JNUIntMethod(
		getClass()
		.getMethod("capacity", "()I")
		).invoke(_object);
}

jsize	JNUBuffer::getLimit() const
	throw (java::jnu::PendingJavaException)
{
	return JNUIntMethod(
		getClass()
		.getMethod("limit", "()I")
		).invoke(_object);
}

jsize	JNUBuffer::getPosition() const
	throw (java::jnu::PendingJavaException)
{
	return JNUIntMethod(
		getClass()
		.getMethod("position", "()I")
		).invoke(_object);
}

void	JNUBuffer::setLimit(jsize limit)
	throw (java::jnu::PendingJavaException)
{
	JNUObjectMethod(
		getClass()
		.getMethod("limit", "(I)Ljava/nio/Buffer;")
		).invoke(_object, limit);
}

void	JNUBuffer::setPosition(jsize position)
	throw (java::jnu::PendingJavaException)
{
	JNUObjectMethod(
		getClass()
		.getMethod("position", "(I)Ljava/nio/Buffer;")
		).invoke(_object, position);
}

// JNUByteBuffer
JNUByteBuffer::JNUByteBuffer(void *address, jsize capacity, jint mode)
	throw (java::jnu::PendingJavaException)
	: JNUBuffer(JNUJVM::env()->NewDirectByteBuffer(address, capacity)),
	_data(address),
	_capacity(capacity),
	_array(0),
	_isDirect(JNI_TRUE)
{
	JNUEnv	env = JNUJVM::env();
//	printf("JNUByteBuffer::JNUByteBuffer(%x, %d, %d) env=%x _object=%x\n", address, capacity, mode, (JNIEnv*)env, _object); fflush(stdout);
	env.checkPendingException();
	if (_object == 0)
	{
		_isDirect = JNI_FALSE;
		// This JVM can't access a direct BB.
		JNUClass cls = JNUClass::forName("java/nio/ByteBuffer", _weak);
		_object = JNUObjectMethod(
			cls.getMethod("allocate", "(I)Ljava/nio/ByteBuffer;")
			).invoke(cls, capacity);
		JNUObject<jbyteArray>	byteArray = (jbyteArray)JNUObjectMethod(
			cls.getMethod("array", "()[B")
			).invoke(_object);

//		jboolean	ok = cls.isInstance(byteArray);

		_array = new JNUByteArray(byteArray, mode);
		_data = const_cast<jbyte*>(_array->getElements());
		_capacity = _array->length;
		memcpy(_data, address, (unsigned)capacity);
	}
}

JNUByteBuffer::JNUByteBuffer(const JNUObject<jobject> &buffer, jint mode)
	throw (java::jnu::PendingJavaException)
	: JNUBuffer(buffer),
	_data(JNUJVM::env()->GetDirectBufferAddress(buffer)),
	_capacity(JNUJVM::env()->GetDirectBufferCapacity(buffer)),
	_array(0),
	_isDirect(JNI_TRUE)
{
	//		fprintf(stderr, "JNUByteBuffer: env=%8.8x bb=%8.8x\n", env, bb);
	//		fprintf(stderr, "JNUByteBuffer: _data=%8.8x\n", _data);
	JNUJVM::env().checkPendingException();
	if (_data == 0)
	{
		_isDirect = JNI_FALSE;
		// Not a direct BB, or this JVM can't access a direct BB.
		JNUClass cls = getClass();
		JNUObject<jbyteArray>	byteArray = (jbyteArray)JNUObjectMethod(
			cls.getMethod("array", "()[B")
			)
			.invoke(_object);

//		jboolean	ok = cls.isInstance(byteArray);

		_array = new JNUByteArray(byteArray, mode);
		_data = const_cast<jbyte*>(_array->getElements());
		_capacity = _array->length;
	}
	//		fprintf(stderr, "JNUByteBuffer: data=%8.8x length=%d\n", _data, _length);
}

JNUByteBuffer::~JNUByteBuffer() throw()
{
	delete _array;
}

JNUByteBuffer	JNUByteBuffer::wrap(void *address, jsize capacity, jint mode)
	 throw(PendingJavaException)
{
	return JNUByteBuffer(address, capacity, mode);
}

void	JNUByteBuffer::commit() throw(PendingJavaException)
{
	if (_array)
		_array->commit();
	// else buffer is direct so changes to data are already committed.
}



jweak JNUByteBuffer::_weak;

// Method pointers

// JNUField

template<> jboolean (JNIEnv::*const JNUBooleanField::GetField)(jobject object, jfieldID fieldID)					= &JNIEnv::GetBooleanField;
template<> jboolean (JNIEnv::*const JNUBooleanField::GetStaticField)(jclass clazz, jfieldID fieldID)				= &JNIEnv::GetStaticBooleanField;
template<> void (JNIEnv::*const JNUBooleanField::SetField)(jobject object, jfieldID fieldID, jboolean value)		= &JNIEnv::SetBooleanField;
template<> void (JNIEnv::*const JNUBooleanField::SetStaticField)(jclass clazz, jfieldID fieldID, jboolean value)	= &JNIEnv::SetStaticBooleanField;

template<> jbyte (JNIEnv::*const JNUByteField::GetField)(jobject object, jfieldID fieldID)					= &JNIEnv::GetByteField;
template<> jbyte (JNIEnv::*const JNUByteField::GetStaticField)(jclass clazz, jfieldID fieldID)				= &JNIEnv::GetStaticByteField;
template<> void (JNIEnv::*const JNUByteField::SetField)(jobject object, jfieldID fieldID, jbyte value)		= &JNIEnv::SetByteField;
template<> void (JNIEnv::*const JNUByteField::SetStaticField)(jclass clazz, jfieldID fieldID, jbyte value)	= &JNIEnv::SetStaticByteField;

template<> jchar (JNIEnv::*const JNUCharField::GetField)(jobject object, jfieldID fieldID)					= &JNIEnv::GetCharField;
template<> jchar (JNIEnv::*const JNUCharField::GetStaticField)(jclass clazz, jfieldID fieldID)				= &JNIEnv::GetStaticCharField;
template<> void (JNIEnv::*const JNUCharField::SetField)(jobject object, jfieldID fieldID, jchar value)		= &JNIEnv::SetCharField;
template<> void (JNIEnv::*const JNUCharField::SetStaticField)(jclass clazz, jfieldID fieldID, jchar value)	= &JNIEnv::SetStaticCharField;

template<> jdouble (JNIEnv::*const JNUDoubleField::GetField)(jobject object, jfieldID fieldID)					= &JNIEnv::GetDoubleField;
template<> jdouble (JNIEnv::*const JNUDoubleField::GetStaticField)(jclass clazz, jfieldID fieldID)				= &JNIEnv::GetStaticDoubleField;
template<> void (JNIEnv::*const JNUDoubleField::SetField)(jobject object, jfieldID fieldID, jdouble value)		= &JNIEnv::SetDoubleField;
template<> void (JNIEnv::*const JNUDoubleField::SetStaticField)(jclass clazz, jfieldID fieldID, jdouble value)	= &JNIEnv::SetStaticDoubleField;

template<> jfloat (JNIEnv::*const JNUFloatField::GetField)(jobject object, jfieldID fieldID)					= &JNIEnv::GetFloatField;
template<> jfloat (JNIEnv::*const JNUFloatField::GetStaticField)(jclass clazz, jfieldID fieldID)				= &JNIEnv::GetStaticFloatField;
template<> void (JNIEnv::*const JNUFloatField::SetField)(jobject object, jfieldID fieldID, jfloat value)		= &JNIEnv::SetFloatField;
template<> void (JNIEnv::*const JNUFloatField::SetStaticField)(jclass clazz, jfieldID fieldID, jfloat value)	= &JNIEnv::SetStaticFloatField;

template<> jint (JNIEnv::*const JNUIntField::GetField)(jobject object, jfieldID fieldID)					= &JNIEnv::GetIntField;
template<> jint (JNIEnv::*const JNUIntField::GetStaticField)(jclass clazz, jfieldID fieldID)				= &JNIEnv::GetStaticIntField;
template<> void (JNIEnv::*const JNUIntField::SetField)(jobject object, jfieldID fieldID, jint value)		= &JNIEnv::SetIntField;
template<> void (JNIEnv::*const JNUIntField::SetStaticField)(jclass clazz, jfieldID fieldID, jint value)	= &JNIEnv::SetStaticIntField;

template<> jlong (JNIEnv::*const JNULongField::GetField)(jobject object, jfieldID fieldID)					= &JNIEnv::GetLongField;
template<> jlong (JNIEnv::*const JNULongField::GetStaticField)(jclass clazz, jfieldID fieldID)				= &JNIEnv::GetStaticLongField;
template<> void (JNIEnv::*const JNULongField::SetField)(jobject object, jfieldID fieldID, jlong value)		= &JNIEnv::SetLongField;
template<> void (JNIEnv::*const JNULongField::SetStaticField)(jclass clazz, jfieldID fieldID, jlong value)	= &JNIEnv::SetStaticLongField;

template<> jshort (JNIEnv::*const JNUShortField::GetField)(jobject object, jfieldID fieldID)					= &JNIEnv::GetShortField;
template<> jshort (JNIEnv::*const JNUShortField::GetStaticField)(jclass clazz, jfieldID fieldID)				= &JNIEnv::GetStaticShortField;
template<> void (JNIEnv::*const JNUShortField::SetField)(jobject object, jfieldID fieldID, jshort value)		= &JNIEnv::SetShortField;
template<> void (JNIEnv::*const JNUShortField::SetStaticField)(jclass clazz, jfieldID fieldID, jshort value)	= &JNIEnv::SetStaticShortField;

template<> jobject (JNIEnv::*const JNUObjectField::GetField)(jobject object, jfieldID fieldID)					= &JNIEnv::GetObjectField;
template<> jobject (JNIEnv::*const JNUObjectField::GetStaticField)(jclass clazz, jfieldID fieldID)				= &JNIEnv::GetStaticObjectField;
template<> void (JNIEnv::*const JNUObjectField::SetField)(jobject object, jfieldID fieldID, jobject value)		= &JNIEnv::SetObjectField;
template<> void (JNIEnv::*const JNUObjectField::SetStaticField)(jclass clazz, jfieldID fieldID, jobject value)	= &JNIEnv::SetStaticObjectField;

// JNUMethod

// Method pointers

// template<> jboolean (JNIEnv::*const JNUBooleanMethod::CallMethod)(jobject, jmethodID, ...)						= &JNIEnv::CallBooleanMethod;
template<> jboolean (JNIEnv::*const JNUBooleanMethod::CallMethodA)(jobject, jmethodID, const jvalue *)			= &JNIEnv::CallBooleanMethodA;
template<> jboolean (JNIEnv::*const JNUBooleanMethod::CallMethodV)(jobject, jmethodID, va_list)					= &JNIEnv::CallBooleanMethodV;
// template<> jboolean (JNIEnv::*const JNUBooleanMethod::CallStaticMethod)(jclass, jmethodID, ...)					= &JNIEnv::CallStaticBooleanMethod;
template<> jboolean (JNIEnv::*const JNUBooleanMethod::CallStaticMethodA)(jclass, jmethodID, const jvalue *)		= &JNIEnv::CallStaticBooleanMethodA;
template<> jboolean (JNIEnv::*const JNUBooleanMethod::CallStaticMethodV)(jclass, jmethodID, va_list)				= &JNIEnv::CallStaticBooleanMethodV;
// template<> jboolean (JNIEnv::*const JNUBooleanMethod::CallNonvirtualMethod)(jobject, jclass, jmethodID, ...)		= &JNIEnv::CallNonvirtualBooleanMethod;
template<> jboolean (JNIEnv::*const JNUBooleanMethod::CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *) = &JNIEnv::CallNonvirtualBooleanMethodA;
template<> jboolean (JNIEnv::*const JNUBooleanMethod::CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list) = &JNIEnv::CallNonvirtualBooleanMethodV;

// template<> jbyte (JNIEnv::*const JNUByteMethod::CallMethod)(jobject, jmethodID, ...)								= &JNIEnv::CallByteMethod;
template<> jbyte (JNIEnv::*const JNUByteMethod::CallMethodA)(jobject, jmethodID, const jvalue *)					= &JNIEnv::CallByteMethodA;
template<> jbyte (JNIEnv::*const JNUByteMethod::CallMethodV)(jobject, jmethodID, va_list) 						= &JNIEnv::CallByteMethodV;
// template<> jbyte (JNIEnv::*const JNUByteMethod::CallStaticMethod)(jclass, jmethodID, ...) 						= &JNIEnv::CallStaticByteMethod;
template<> jbyte (JNIEnv::*const JNUByteMethod::CallStaticMethodA)(jclass, jmethodID, const jvalue *) 			= &JNIEnv::CallStaticByteMethodA;
template<> jbyte (JNIEnv::*const JNUByteMethod::CallStaticMethodV)(jclass, jmethodID, va_list) 					= &JNIEnv::CallStaticByteMethodV;
// template<> jbyte (JNIEnv::*const JNUByteMethod::CallNonvirtualMethod)(jobject, jclass, jmethodID, ...) 			= &JNIEnv::CallNonvirtualByteMethod;
template<> jbyte (JNIEnv::*const JNUByteMethod::CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *) = &JNIEnv::CallNonvirtualByteMethodA;
template<> jbyte (JNIEnv::*const JNUByteMethod::CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list) 		= &JNIEnv::CallNonvirtualByteMethodV;

// template<> jchar (JNIEnv::*const JNUCharMethod::CallMethod)(jobject, jmethodID, ...) 							= &JNIEnv::CallCharMethod;
template<> jchar (JNIEnv::*const JNUCharMethod::CallMethodA)(jobject, jmethodID, const jvalue *) 				= &JNIEnv::CallCharMethodA;
template<> jchar (JNIEnv::*const JNUCharMethod::CallMethodV)(jobject, jmethodID, va_list) 						= &JNIEnv::CallCharMethodV;
// template<> jchar (JNIEnv::*const JNUCharMethod::CallStaticMethod)(jclass, jmethodID, ...) 						= &JNIEnv::CallStaticCharMethod;
template<> jchar (JNIEnv::*const JNUCharMethod::CallStaticMethodA)(jclass, jmethodID, const jvalue *) 			= &JNIEnv::CallStaticCharMethodA;
template<> jchar (JNIEnv::*const JNUCharMethod::CallStaticMethodV)(jclass, jmethodID, va_list) 					= &JNIEnv::CallStaticCharMethodV;
// template<> jchar (JNIEnv::*const JNUCharMethod::CallNonvirtualMethod)(jobject, jclass, jmethodID, ...) 			= &JNIEnv::CallNonvirtualCharMethod;
template<> jchar (JNIEnv::*const JNUCharMethod::CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *) = &JNIEnv::CallNonvirtualCharMethodA;
template<> jchar (JNIEnv::*const JNUCharMethod::CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list) 		= &JNIEnv::CallNonvirtualCharMethodV;

// template<> jdouble (JNIEnv::*const JNUDoubleMethod::CallMethod)(jobject, jmethodID, ...) 						= &JNIEnv::CallDoubleMethod;
template<> jdouble (JNIEnv::*const JNUDoubleMethod::CallMethodA)(jobject, jmethodID, const jvalue *) 			= &JNIEnv::CallDoubleMethodA;
template<> jdouble (JNIEnv::*const JNUDoubleMethod::CallMethodV)(jobject, jmethodID, va_list) 					= &JNIEnv::CallDoubleMethodV;
// template<> jdouble (JNIEnv::*const JNUDoubleMethod::CallStaticMethod)(jclass, jmethodID, ...) 					= &JNIEnv::CallStaticDoubleMethod;
template<> jdouble (JNIEnv::*const JNUDoubleMethod::CallStaticMethodA)(jclass, jmethodID, const jvalue *) 		= &JNIEnv::CallStaticDoubleMethodA;
template<> jdouble (JNIEnv::*const JNUDoubleMethod::CallStaticMethodV)(jclass, jmethodID, va_list) 				= &JNIEnv::CallStaticDoubleMethodV;
// template<> jdouble (JNIEnv::*const JNUDoubleMethod::CallNonvirtualMethod)(jobject, jclass, jmethodID, ...) 		= &JNIEnv::CallNonvirtualDoubleMethod;
template<> jdouble (JNIEnv::*const JNUDoubleMethod::CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *) = &JNIEnv::CallNonvirtualDoubleMethodA;
template<> jdouble (JNIEnv::*const JNUDoubleMethod::CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list)	= &JNIEnv::CallNonvirtualDoubleMethodV;

// template<> jfloat (JNIEnv::*const JNUFloatMethod::CallMethod)(jobject, jmethodID, ...) 							= &JNIEnv::CallFloatMethod;
template<> jfloat (JNIEnv::*const JNUFloatMethod::CallMethodA)(jobject, jmethodID, const jvalue *) 				= &JNIEnv::CallFloatMethodA;
template<> jfloat (JNIEnv::*const JNUFloatMethod::CallMethodV)(jobject, jmethodID, va_list) 						= &JNIEnv::CallFloatMethodV;
// template<> jfloat (JNIEnv::*const JNUFloatMethod::CallStaticMethod)(jclass, jmethodID, ...) 						= &JNIEnv::CallStaticFloatMethod;
template<> jfloat (JNIEnv::*const JNUFloatMethod::CallStaticMethodA)(jclass, jmethodID, const jvalue *) 			= &JNIEnv::CallStaticFloatMethodA;
template<> jfloat (JNIEnv::*const JNUFloatMethod::CallStaticMethodV)(jclass, jmethodID, va_list) 				= &JNIEnv::CallStaticFloatMethodV;
// template<> jfloat (JNIEnv::*const JNUFloatMethod::CallNonvirtualMethod)(jobject, jclass, jmethodID, ...) 		= &JNIEnv::CallNonvirtualFloatMethod;
template<> jfloat (JNIEnv::*const JNUFloatMethod::CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *) = &JNIEnv::CallNonvirtualFloatMethodA;
template<> jfloat (JNIEnv::*const JNUFloatMethod::CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list)	= &JNIEnv::CallNonvirtualFloatMethodV;

// template<> jint (JNIEnv::*const JNUIntMethod::CallMethod)(jobject, jmethodID, ...) 								= &JNIEnv::CallIntMethod;
template<> jint (JNIEnv::*const JNUIntMethod::CallMethodA)(jobject, jmethodID, const jvalue *) 					= &JNIEnv::CallIntMethodA;
template<> jint (JNIEnv::*const JNUIntMethod::CallMethodV)(jobject, jmethodID, va_list) 							= &JNIEnv::CallIntMethodV;
// template<> jint (JNIEnv::*const JNUIntMethod::CallStaticMethod)(jclass, jmethodID, ...) 							= &JNIEnv::CallStaticIntMethod;
template<> jint (JNIEnv::*const JNUIntMethod::CallStaticMethodA)(jclass, jmethodID, const jvalue *) 				= &JNIEnv::CallStaticIntMethodA;
template<> jint (JNIEnv::*const JNUIntMethod::CallStaticMethodV)(jclass, jmethodID, va_list) 					= &JNIEnv::CallStaticIntMethodV;
// template<> jint (JNIEnv::*const JNUIntMethod::CallNonvirtualMethod)(jobject, jclass, jmethodID, ...) 			= &JNIEnv::CallNonvirtualIntMethod;
template<> jint (JNIEnv::*const JNUIntMethod::CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *) = &JNIEnv::CallNonvirtualIntMethodA;
template<> jint (JNIEnv::*const JNUIntMethod::CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list) 		= &JNIEnv::CallNonvirtualIntMethodV;

// template<> jlong (JNIEnv::*const JNULongMethod::CallMethod)(jobject, jmethodID, ...)							= &JNIEnv::CallLongMethod;
template<> jlong (JNIEnv::*const JNULongMethod::CallMethodA)(jobject, jmethodID, const jvalue *)				= &JNIEnv::CallLongMethodA;
template<> jlong (JNIEnv::*const JNULongMethod::CallMethodV)(jobject, jmethodID, va_list)					= &JNIEnv::CallLongMethodV;
// template<> jlong (JNIEnv::*const JNULongMethod::CallStaticMethod)(jclass, jmethodID, ...)					= &JNIEnv::CallStaticLongMethod;
template<> jlong (JNIEnv::*const JNULongMethod::CallStaticMethodA)(jclass, jmethodID, const jvalue *)		= &JNIEnv::CallStaticLongMethodA;
template<> jlong (JNIEnv::*const JNULongMethod::CallStaticMethodV)(jclass, jmethodID, va_list)				= &JNIEnv::CallStaticLongMethodV;
// template<> jlong (JNIEnv::*const JNULongMethod::CallNonvirtualMethod)(jobject, jclass, jmethodID, ...)		= &JNIEnv::CallNonvirtualLongMethod;
template<> jlong (JNIEnv::*const JNULongMethod::CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *) = &JNIEnv::CallNonvirtualLongMethodA;
template<> jlong (JNIEnv::*const JNULongMethod::CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list)	= &JNIEnv::CallNonvirtualLongMethodV;

// template<> jshort (JNIEnv::*const JNUShortMethod::CallMethod)(jobject, jmethodID, ...) 						= &JNIEnv::CallShortMethod;
template<> jshort (JNIEnv::*const JNUShortMethod::CallMethodA)(jobject, jmethodID, const jvalue *) 			= &JNIEnv::CallShortMethodA;
template<> jshort (JNIEnv::*const JNUShortMethod::CallMethodV)(jobject, jmethodID, va_list) 					= &JNIEnv::CallShortMethodV;
// template<> jshort (JNIEnv::*const JNUShortMethod::CallStaticMethod)(jclass, jmethodID, ...) 					= &JNIEnv::CallStaticShortMethod;
template<> jshort (JNIEnv::*const JNUShortMethod::CallStaticMethodA)(jclass, jmethodID, const jvalue *) 		= &JNIEnv::CallStaticShortMethodA;
template<> jshort (JNIEnv::*const JNUShortMethod::CallStaticMethodV)(jclass, jmethodID, va_list) 			= &JNIEnv::CallStaticShortMethodV;
// template<> jshort (JNIEnv::*const JNUShortMethod::CallNonvirtualMethod)(jobject, jclass, jmethodID, ...) 	= &JNIEnv::CallNonvirtualShortMethod;
template<> jshort (JNIEnv::*const JNUShortMethod::CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *) = &JNIEnv::CallNonvirtualShortMethodA;
template<> jshort (JNIEnv::*const JNUShortMethod::CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list) = &JNIEnv::CallNonvirtualShortMethodV;

/*
 * UNUSED
// template<> void (JNIEnv::*const JNUVoidMethod::CallMethod)(jobject, jmethodID, ...) 							= &JNIEnv::CallVoidMethod;
template<> void (JNIEnv::*const JNUVoidMethod::CallMethodA)(jobject, jmethodID, const jvalue *) 				= &JNIEnv::CallVoidMethodA;
template<> void (JNIEnv::*const JNUVoidMethod::CallMethodV)(jobject, jmethodID, va_list) 					= &JNIEnv::CallVoidMethodV;
// template<> void (JNIEnv::*const JNUVoidMethod::CallStaticMethod)(jclass, jmethodID, ...) 					= &JNIEnv::CallStaticVoidMethod;
template<> void (JNIEnv::*const JNUVoidMethod::CallStaticMethodA)(jclass, jmethodID, const jvalue *) 		= &JNIEnv::CallStaticVoidMethodA;
template<> void (JNIEnv::*const JNUVoidMethod::CallStaticMethodV)(jclass, jmethodID, va_list) 				= &JNIEnv::CallStaticVoidMethodV;
// template<> void (JNIEnv::*const JNUVoidMethod::CallNonvirtualMethod)(jobject, jclass, jmethodID, ...) 		= &JNIEnv::CallNonvirtualVoidMethod;
template<> void (JNIEnv::*const JNUVoidMethod::CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *) = &JNIEnv::CallNonvirtualVoidMethodA;
template<> void (JNIEnv::*const JNUVoidMethod::CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list) 	= &JNIEnv::CallNonvirtualVoidMethodV;
*/
// template<> jobject (JNIEnv::*const JNUObjectMethod::CallMethod)(jobject, jmethodID, ...) 					= &JNIEnv::CallObjectMethod;
template<> jobject (JNIEnv::*const JNUObjectMethod::CallMethodA)(jobject, jmethodID, const jvalue *) 		= &JNIEnv::CallObjectMethodA;
template<> jobject (JNIEnv::*const JNUObjectMethod::CallMethodV)(jobject, jmethodID, va_list) 				= &JNIEnv::CallObjectMethodV;
// template<> jobject (JNIEnv::*const JNUObjectMethod::CallStaticMethod)(jclass, jmethodID, ...) 				= &JNIEnv::CallStaticObjectMethod;
template<> jobject (JNIEnv::*const JNUObjectMethod::CallStaticMethodA)(jclass, jmethodID, const jvalue *) 	= &JNIEnv::CallStaticObjectMethodA;
template<> jobject (JNIEnv::*const JNUObjectMethod::CallStaticMethodV)(jclass, jmethodID, va_list) = &JNIEnv::CallStaticObjectMethodV;
// template<> jobject (JNIEnv::*const JNUObjectMethod::CallNonvirtualMethod)(jobject, jclass, jmethodID, ...) 	= &JNIEnv::CallNonvirtualObjectMethod;
template<> jobject (JNIEnv::*const JNUObjectMethod::CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *) = &JNIEnv::CallNonvirtualObjectMethodA;
template<> jobject (JNIEnv::*const JNUObjectMethod::CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list) = &JNIEnv::CallNonvirtualObjectMethodV;

// JNUPrimitiveArray

// Method pointers

template<> jbooleanArray (JNIEnv::*const JNUBooleanArray::NewArray)(jsize)								= &JNIEnv::NewBooleanArray;
template<> jboolean *(JNIEnv::*const JNUBooleanArray::GetArrayElements)(jbooleanArray, jboolean *)		= &JNIEnv::GetBooleanArrayElements;
template<> void (JNIEnv::*const JNUBooleanArray::ReleaseArrayElements)(jbooleanArray, jboolean *, jint)= &JNIEnv::ReleaseBooleanArrayElements;
template<> void (JNIEnv::*const JNUBooleanArray::GetArrayRegion)(jbooleanArray, jsize, jsize, jboolean *)	= &JNIEnv::GetBooleanArrayRegion;
template<> void (JNIEnv::*const JNUBooleanArray::SetArrayRegion)(jbooleanArray, jsize, jsize, const jboolean *)	= &JNIEnv::SetBooleanArrayRegion;

template<> jbyteArray (JNIEnv::*const JNUByteArray::NewArray)(jsize)									= &JNIEnv::NewByteArray;
template<> jbyte *(JNIEnv::*const JNUByteArray::GetArrayElements)(jbyteArray, jboolean *)					= &JNIEnv::GetByteArrayElements;
template<> void (JNIEnv::*const JNUByteArray::ReleaseArrayElements)(jbyteArray, jbyte *, jint)			= &JNIEnv::ReleaseByteArrayElements;
template<> void (JNIEnv::*const JNUByteArray::GetArrayRegion)(jbyteArray, jsize, jsize, jbyte *)		= &JNIEnv::GetByteArrayRegion;
template<> void (JNIEnv::*const JNUByteArray::SetArrayRegion)(jbyteArray, jsize, jsize, const jbyte *)	= &JNIEnv::SetByteArrayRegion;

template<> jcharArray (JNIEnv::*const JNUCharArray::NewArray)(jsize)									= &JNIEnv::NewCharArray;
template<> jchar *(JNIEnv::*const JNUCharArray::GetArrayElements)(jcharArray, jboolean *)					= &JNIEnv::GetCharArrayElements;
template<> void (JNIEnv::*const JNUCharArray::ReleaseArrayElements)(jcharArray, jchar *, jint)			= &JNIEnv::ReleaseCharArrayElements;
template<> void (JNIEnv::*const JNUCharArray::GetArrayRegion)(jcharArray, jsize, jsize, jchar *)		= &JNIEnv::GetCharArrayRegion;
template<> void (JNIEnv::*const JNUCharArray::SetArrayRegion)(jcharArray, jsize, jsize, const jchar *)	= &JNIEnv::SetCharArrayRegion;

template<> jdoubleArray (JNIEnv::*const JNUDoubleArray::NewArray)(jsize)								= &JNIEnv::NewDoubleArray;
template<> jdouble *(JNIEnv::*const JNUDoubleArray::GetArrayElements)(jdoubleArray, jboolean *)			= &JNIEnv::GetDoubleArrayElements;
template<> void (JNIEnv::*const JNUDoubleArray::ReleaseArrayElements)(jdoubleArray, jdouble *, jint)	= &JNIEnv::ReleaseDoubleArrayElements;
template<> void (JNIEnv::*const JNUDoubleArray::GetArrayRegion)(jdoubleArray, jsize, jsize, jdouble *)		= &JNIEnv::GetDoubleArrayRegion;
template<> void (JNIEnv::*const JNUDoubleArray::SetArrayRegion)(jdoubleArray, jsize, jsize, const jdouble *)	= &JNIEnv::SetDoubleArrayRegion;

template<> jfloatArray (JNIEnv::*const JNUFloatArray::NewArray)(jsize)									= &JNIEnv::NewFloatArray;
template<> jfloat *(JNIEnv::*const JNUFloatArray::GetArrayElements)(jfloatArray, jboolean *)				= &JNIEnv::GetFloatArrayElements;
template<> void (JNIEnv::*const JNUFloatArray::ReleaseArrayElements)(jfloatArray, jfloat *, jint)		= &JNIEnv::ReleaseFloatArrayElements;
template<> void (JNIEnv::*const JNUFloatArray::GetArrayRegion)(jfloatArray, jsize, jsize, jfloat *)		= &JNIEnv::GetFloatArrayRegion;
template<> void (JNIEnv::*const JNUFloatArray::SetArrayRegion)(jfloatArray, jsize, jsize, const jfloat *)	= &JNIEnv::SetFloatArrayRegion;

template<> jintArray (JNIEnv::*const JNUIntArray::NewArray)(jsize)										= &JNIEnv::NewIntArray;
template<> jint *(JNIEnv::*const JNUIntArray::GetArrayElements)(jintArray, jboolean *)						= &JNIEnv::GetIntArrayElements;
template<> void (JNIEnv::*const JNUIntArray::ReleaseArrayElements)(jintArray, jint *, jint)			= &JNIEnv::ReleaseIntArrayElements;
template<> void (JNIEnv::*const JNUIntArray::GetArrayRegion)(jintArray, jsize, jsize, jint *)			= &JNIEnv::GetIntArrayRegion;
template<> void (JNIEnv::*const JNUIntArray::SetArrayRegion)(jintArray, jsize, jsize, const jint *)	= &JNIEnv::SetIntArrayRegion;

template<> jlongArray (JNIEnv::*const JNULongArray::NewArray)(jsize)									= &JNIEnv::NewLongArray;
template<> jlong *(JNIEnv::*const JNULongArray::GetArrayElements)(jlongArray, jboolean *)					= &JNIEnv::GetLongArrayElements;
template<> void (JNIEnv::*const JNULongArray::ReleaseArrayElements)(jlongArray, jlong *, jint)			= &JNIEnv::ReleaseLongArrayElements;
template<> void (JNIEnv::*const JNULongArray::GetArrayRegion)(jlongArray, jsize, jsize, jlong *)		= &JNIEnv::GetLongArrayRegion;
template<> void (JNIEnv::*const JNULongArray::SetArrayRegion)(jlongArray, jsize, jsize, const jlong *)		= &JNIEnv::SetLongArrayRegion;

template<> jshortArray (JNIEnv::*const JNUShortArray::NewArray)(jsize)									= &JNIEnv::NewShortArray;
template<> jshort *(JNIEnv::*const JNUShortArray::GetArrayElements)(jshortArray, jboolean *)				= &JNIEnv::GetShortArrayElements;
template<> void (JNIEnv::*const JNUShortArray::ReleaseArrayElements)(jshortArray, jshort *, jint)		= &JNIEnv::ReleaseShortArrayElements;
template<> void (JNIEnv::*const JNUShortArray::GetArrayRegion)(jshortArray, jsize, jsize, jshort *)		= &JNIEnv::GetShortArrayRegion;
template<> void (JNIEnv::*const JNUShortArray::SetArrayRegion)(jshortArray, jsize, jsize, const jshort *)	= &JNIEnv::SetShortArrayRegion;

#if COMPILATION_TEST
// Exercise all templates.
void	dummy(JNIEnv *env, jobject object, jclass cls)
{
	JNUObject<jobject>	jnuObject(object);
	JNUClass			jnuClass(cls);
	JNUFieldID			fid = jnuClass.getField("name", "signature");
	JNUMethodID			mid = jnuClass.getMethod("name", "signature");

	JNUBooleanField	booleanField(fid);
	booleanField.set(jnuObject, booleanField.get(jnuObject));
	JNUByteField	byteField(fid);
	byteField.set(jnuObject, byteField.get(jnuObject));
	JNUCharField	charField(fid);
	charField.set(jnuObject, charField.get(jnuObject));
	JNUDoubleField	doubleField(fid);
	doubleField.set(jnuObject, doubleField.get(jnuObject));
	JNUFloatField	floatField(fid);
	floatField.set(jnuObject, floatField.get(jnuObject));
	JNUIntField		intField(fid);
	intField.set(jnuObject, intField.get(jnuObject));
	JNULongField	longField(fid);
	longField.set(jnuObject, longField.get(jnuObject));
	JNUShortField	shortField(fid);
	shortField.set(jnuObject, shortField.get(jnuObject));
	JNUObjectField	objectField(fid);
	objectField.set(jnuObject, objectField.get(jnuObject));

	JNUBooleanMethod	booleanMethod(mid);
	booleanMethod.invoke(jnuObject);
	JNUByteMethod	byteMethod(mid);
	byteMethod.invoke(jnuObject);
	JNUCharMethod	charMethod(mid);
	charMethod.invoke(jnuObject);
	JNUDoubleMethod	doubleMethod(mid);
	doubleMethod.invoke(jnuObject);
	JNUFloatMethod	floatMethod(mid);	
	floatMethod.invoke(jnuObject);
	JNUIntMethod	intMethod(mid);
	intMethod.invoke(jnuObject);
	JNULongMethod	longMethod(mid);
	longMethod.invoke(jnuObject);
	JNUShortMethod	shortMethod(mid);
	shortMethod.invoke(jnuObject);
	JNUVoidMethod	voidMethod(mid);
	voidMethod.invoke(jnuObject);
	JNUObjectMethod	objectMethod(mid);
	objectMethod.invoke(jnuObject);

	JNUBooleanArray	booleanArray((jbooleanArray)0);
	JNUByteArray	byteArray((jbyteArray)0);
	JNUCharArray	charArray((jcharArray)0);
	JNUDoubleArray	doubleArray((jdoubleArray)0);
	JNUFloatArray	floatArray((jfloatArray)0);
	JNUIntArray		intArray((jintArray)0);
	JNULongArray	longArray((jlongArray)0);
	JNUShortArray	shortArray((jshortArray)0);
	JNUObjectArray	objectArray((jobjectArray)0);

#if NOTYET
	JNUBooleanArrayCritical	booleanArrayCritical((jbooleanArray)0);
	JNUByteArrayCritical	byteArrayCritical((jbyteArray)0);
	JNUCharArrayCritical	charArrayCritical((jcharArray)0);
	JNUDoubleArrayCritical	doubleArrayCritical((jdoubleArray)0);
	JNUFloatArrayCritical	floatArrayCritical((jfloatArray)0);
	JNUIntArrayCritical		intArrayCritical((jintArray)0);
	JNULongArrayCritical	longArrayCritical((jlongArray)0);
	JNUShortArrayCritical	shortArrayCritical((jshortArray)0);
#endif
}
#endif

#ifdef _MSC_VER
#pragma warning(pop)
#endif

