/*
 * CaptureDevice.java
 *
 * Created on 8 May 2008, 14:13
 *
 * Copyright (c) Esmond Pitt, 2008. All rights reserved.
 */

package net.sourceforge.jpcap.capture;
import java.net.InetAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.UnknownHostException;
import java.util.Arrays;


/**
 * Encapsulates a capture-capable device.
 *<p>
 * NB there is no support for supplying the MAC address of a capture device.
 * In Java 6, you can get it this way:
 * <pre><code>
 * NetworkInterface nwif = NetworkInterface.getByInetAddress(captureDevice.getAddresses()[0].getAddress());
 * byte[] macAddress = nwif.getHardwareAddress();
 *</code></pre>
 *
 * @author Esmond Pitt
 * @version $Revision: 1.1 $
 * @since 2.0.0
 */
public class CaptureDevice
{
	/**
	 * Loopback bitmask. Must match the declaration in pcap.h.
	 */
	public static final int	PCAP_IF_LOOPBACK = 0x00000001;
	
	private final String name;
	private final String description;
	private int	network;
	private int	netmask;
	private int flags;
//	private byte[]	macAddress = new byte[6];
	private Address[]	addresses;
	private final boolean loopback;

	/**
	 * Construct a new instance of CaptureDevice.
	 * This constructor is used by JNI code.
	 *
	 * @param name Device name.
	 * @param description Device description if available, or null.
	 * @param network Network, from pcap_lookupnet(), corrected to host byte order
	 * @param netmask Netmask, from pcap_lookupnet(), corrected to host byte order
	 * @param flags pcap_if_t.flags.
	 * @param addresses Addresses of this interface
	 */
	protected CaptureDevice(String name, String description, int network, int netmask, int flags, Address[] addresses)
	{
		this.name = name;
		this.description = description;
		this.network = network;
		this.netmask = netmask;
		this.flags = flags;
		this.addresses = addresses;
		this.loopback = (flags & PCAP_IF_LOOPBACK) != 0;
//		System.out.println("CaptureDevice.<init>(name="+name+",description="+description+",network="+Integer.toHexString(network)+",netmask="+Integer.toHexString(netmask)+",flags="+flags+")");
		if (addresses == null || addresses.length < 1)
			throw new IllegalArgumentException(this.toString()+": device is down, no IP addresses");
	}
	
	public String	toString()
	{
		return "CaptureDevice "
			+name
			+":"+description
			// +":mac="+MACAddress.extract(0,  macAddress)
			+":loopback="+loopback
			+":addresses="+Arrays.asList(addresses);
	}

	/**
	 * Get the device name. This is the attribute that is passed to {@link PacketCaptureCapable#open}.
	 * @return the device name.
	 */
	public String getName()
	{
		return this.name;
	}

	/**
	 * Get the device description.
	 * @return the device description.
	 */
	public String getDescription()
	{
		return this.description;
	}
	
	/**
	 * Return the MAC address.
	 * @return the MAC address
	public byte[]	getMACAddress()
	{
		return macAddress;
	}
	 */
	
	/**
	 * Set the MAC address.
	 * This is package-private: this class is not intended to be mutable by the application.
	 * @param macAddress MAC address
	void	setMACAddress(byte[] macAddress)
	{
		this.macAddress = macAddress;
	}
	 */

	/**
	 * Return the netmask.
	 * This method is package-private, as it is only required for the purposes of
	 * {@link PacketCapture#setFilter}.
	 * The netmask returned by this method is IPv4 only and doesn't account
	 * for the possibility that the device has multiple addresses and therefore possibly
	 * multiple netmasks, so it's not of general utility.
	 * @return the netmask.
	 */
	int	getNetmask()
	{
		return netmask;
	}

	/**
	 * Return the network.
	 * This method is package-private, as it is only required for the purposes of
	 * {@link PacketCapture#getNetwork}, whic is itself deprecated.
	 * The network returned by this method is IPv4 only and doesn't account
	 * for the possibility that the device has multiple addresses and therefore possibly
	 * multiple netmasks, so it's not of general utility.
	 * @return the netmask.
	 */
	int	getNetwork()
	{
		return network;
	}
	
	/**
	 * Return true if the device is a loopback device.
	 * @return true if the device is a loopback device.
	 */
	public boolean isLoopback()
	{
		return this.loopback;
	}
	
	/**
	 * Return the address information for the device.
	 * @return the addressing information for the device.
	 */
	public Address[]	getAddresses()
	{
		return addresses;
	}
	
	public int	hashCode()
	{
		return name.hashCode()
		^ network
		^ netmask
		^ flags
		^ Arrays.hashCode(addresses)
		;
	}
	
	public boolean	equals(Object other)
	{
		if (other instanceof CaptureDevice)
		{
			CaptureDevice	that = (CaptureDevice)other;
			return this.name.equals(that.name)
			&& this.network == that.network
			&& this.netmask == that.netmask
			&& this.flags == that.flags
			&& Arrays.equals(this.addresses, that.addresses)
			;
		}
		return false;
	}
	
	/**
	 * The Address class encapsulates addressing information for a CaptureDevice.
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 * @since 2.0.0
	 */
	public static class Address
	{
		private InetAddress address;
		private InetAddress netmask;
		private InetAddress broadcast;
		private InetAddress destination;
	
		/**
		 * Construct an Address.
		 * This constructor is used by JNI code.
		 * @param address IP address
		 * @param netmask Netmask
		 * @param broadcast Broadcast address
		 * @param destination Destination address
		 */
		Address(byte[] address, byte[] netmask, byte[] broadcast, byte[] destination)
			throws UnknownHostException
		{
			this.address = InetAddress.getByAddress(address);
			this.netmask = InetAddress.getByAddress(netmask);
			this.broadcast = InetAddress.getByAddress(broadcast);
			this.destination = InetAddress.getByAddress(destination);
		}

		/**
		 * Get the IP address.
		 * @return the IP address.
		 */
		public InetAddress getAddress()
		{
			return this.address;
		}

		/**
		 * Get the netmask.
		 * @return the netmask.
		 */
		public InetAddress getNetmask()
		{
			return this.netmask;
		}

		/**
		 * Get the broadcast address.
		 * @return the broadcast address.
		 */
		public InetAddress getBroadcast()
		{
			return this.broadcast;
		}

		/**
		 * Get the destination address: zero-length if the device isn't a point-to-point device.
		 * @return the destination address.
		 */
		public InetAddress getDestination()
		{
			return this.destination;
		}
		
		public int	hashCode()
		{
			return address.hashCode()
			^ netmask.hashCode()
			^ broadcast.hashCode()
			^ destination.hashCode();
		}
		
		public boolean	equals(Object other)
		{
			if (other instanceof Address)
			{
				Address	that = (Address)other;
				return this.address.equals(that.address)
				&& this.netmask.equals(that.netmask)
				&& this.broadcast.equals(that.broadcast)
				&& this.destination.equals(that.destination)
				;
			}
			return false;
		}
		
		public String	toString()
		{
			StringBuffer	sb = new StringBuffer();
			sb.append("Address[");
			try
			{
				sb.append("address="+address.getHostAddress());
				sb.append(", netmask="+netmask.getHostAddress());
				sb.append(", broadcast="+broadcast.getHostAddress());
				sb.append(", destination="+destination.getHostAddress());
			}
			finally
			{
				sb.append("]");
				return sb.toString();
			}
		}
		
	}
	
}
