/*
 * CDP.java
 *
 * Created on 30 April 2008, 15:55
 *
 ***************************************************************************
 * Copyright (C) 2008, Esmond Pitt                                         *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************
 */

package net.sourceforge.jpcap.net;

import java.nio.ByteBuffer;
import net.sourceforge.jpcap.util.AnsiEscapeSequences;
import net.sourceforge.jpcap.util.Timeval;

/**
 * Cisco Discovery Protocol (CDP).
 * This is an Ethernet 802.3 SNAP protocol.
 *
 * @since 2.0.0
 *
 * @author Esmond Pitt
 * @version $Revision: 1.1 $
 */
public interface CDP extends SNAP
{
	/**
	 * Offset of CDP header
	 */
	public static final int CDP_HEADER_POS = SNAP_HEADER_POS+SNAP_HEADER_LEN;
	
	/**
	 * Offset of CDP version
	 */
	public static final int CDP_VERSION_POS = CDP_HEADER_POS+0;

	/**
	 * Length of CDP version
	 */
	public static final int CDP_VERSION_LEN = 1;

	/**
	 * Offset of CDP time-to-live
	 */
	public static final int CDP_TTL_POS = CDP_VERSION_POS+CDP_VERSION_LEN;

	/**
	 * Length of CDP time-to-live
	 */
	public static final int CDP_TTL_LEN = 1;
	
	/**
	 * Offset of CDP checksum
	 */
	public static final int CDP_CHECKSUM_POS = CDP_TTL_POS+CDP_TTL_LEN;
	
	/**
	 * Length of CDP checksum
	 */
	public static final int CDP_CHECKSUM_LEN = 2;
	
	/**
	 * Length of CDP header
	 */
	public static final int CDP_HEADER_LEN = CDP_VERSION_LEN+CDP_TTL_LEN+CDP_CHECKSUM_LEN;

//	static
//	{
//		System.out.println("CDP.CDP_CHECKSUM_POS="+CDP_CHECKSUM_POS);
//	}

	/**
	 * CDP Packet class.
	 *
	 * @since 2.0.0
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 */
	class Packet extends EthernetPacket 
	{
		static final long serialVersionUID = -5265870644322223513L;
		
		/** Creates a new CDP packet */
		public Packet(int lLen, byte [] bytes)
		{
			this(lLen, bytes, null);
		}
		
		/**
		 * Create a new CDP packet.
		 */
		public Packet(int lLen, byte [] bytes, Timeval tv)
		{
			this(lLen, ByteBuffer.wrap(bytes), tv);
		}
		
		/**
		 * Create a new CDP packet.
		 * @param lLen Packet length
		 * @param bb Packet bytes in a ByteBuffer
		 * @param tv TimeValue
		 */
		public Packet(int lLen, ByteBuffer bb, Timeval tv)
		{
			super(lLen, bb, tv);
		}
		
		/**
		 * @return the CDP protocol version.
		 */
		public int	getCDPVersion()
		{
			return _byteBuffer.get(_ethOffset+CDP_VERSION_POS);
		}
		
		/**
		 * Set the CDP protocol version
		 * @param version CDP protocol version
		 */
		public void	setCDPVersion(int version)
		{
			_byteBuffer.put(_ethOffset+CDP_VERSION_POS, (byte)version);
		}
		
		/**
		 * @return the CDP checksum.
		 */
		public int	getCDPChecksum()
		{
			return _byteBuffer.getShort(_ethOffset+CDP_CHECKSUM_POS) & 0xffff;
		}
		
		/**
		 * @return the CDP time-to-live.
		 */
		public int	getCDPTimeToLive()
		{
			return _byteBuffer.get(_ethOffset+CDP_TTL_POS);
		}
		
		/**
		 * Set the CDP time-to-live (TTL)
		 * @param ttl CDP time-to-live
		 */
		public void	setCDPTimeToLive(int ttl)
		{
			_byteBuffer.put(_ethOffset+CDP_TTL_POS, (byte)ttl);
		}
		
		@Override
			public void	setChecksums()
		{
			// Clear checksum word
			_byteBuffer.putShort(_ethOffset+CDP_CHECKSUM_POS, (short)0);
			int	checksum = super.ipChecksum(toByteArray(), _ethOffset+CDP_HEADER_POS, super.getEthernetData().length-CDP_HEADER_POS);
			_byteBuffer.putShort(_ethOffset+CDP_CHECKSUM_POS, (short)checksum);
			super.setChecksums();
		}
		
		/**
		 * @return a <i>copy</i> of the CDP data, which is a series of MIBs in
		 * the format type:length:value, where 'type' and 'length' are two bytes each and 'length' includes
		 * the length of 'type' and 'length': as the data is a copy, changing it has no effect on the CDP packet object.
		 */
		public byte[]	getCDPData()
		{
//		System.out.println("_ethOffset="+_ethOffset);
//		System.out.println("CDP_HEADER_POS="+CDP_HEADER_POS);
//		System.out.println("CDP_HEADER_LEN="+CDP_HEADER_LEN);
//		System.out.println("CDP data offset="+(_ethOffset+CDP_HEADER_POS+CDP_HEADER_LEN));
//		System.out.println("ethernetData.length="+super.getEthernetData().length);
//		int	headerLen = CDP_HEADER_POS+CDP_HEADER_LEN;
//		System.out.println("headerLen="+headerLen);
//		System.out.println("CDP data length="+(super.getEthernetData().length-headerLen));
//		byte[] data = PacketEncoding.extractData(_ethOffset+CDP_HEADER_POS, CDP_HEADER_LEN, _bytes);
//		System.out.println("extracted data length="+data.length);
//		System.out.println(new java.math.BigInteger(data).toString(16));
//			return PacketEncoding.extractData(_ethOffset+CDP_HEADER_POS, CDP_HEADER_LEN, _bytes);
//			System.out.println("CDP.Packet.getData(): SNAP_HEADER_POS="+SNAP_HEADER_POS);
//			System.out.println("CDP.Packet.getData(): SNAP_HEADER_LEN="+SNAP_HEADER_LEN);
//			System.out.println("CDP.Packet.getData(): CDP_HEADER_POS="+CDP_HEADER_POS);
//			System.out.println("CDP.Packet.getData(): CDP_HEADER_LEN="+CDP_HEADER_LEN);
//			System.out.println("CDP.Packet.getData(): offset="+(_ethOffset+CDP_HEADER_POS+CDP_HEADER_LEN));
			_byteBuffer.position(_ethOffset+CDP_HEADER_POS+CDP_HEADER_LEN);
			 byte[]	data = new byte[_byteBuffer.remaining()];
			 _byteBuffer.get(data);
			 return data;
		}
		
		/**
		 * Set the CDP data payload.
		 * @param data Payload data.
		 */
		public void	setCDPData(byte[] data)
		{
			_byteBuffer.position(_ethOffset+CDP_HEADER_LEN);
			_byteBuffer.put(data);
			_byteBuffer.limit(_byteBuffer.position());
		}
		
		public String	toColoredString(boolean colored)
		{
			StringBuffer buffer = new StringBuffer();
			buffer.append('[');
			if(colored) buffer.append(getColor());
			buffer.append("CDPPacket");
			if(colored) buffer.append(AnsiEscapeSequences.RESET);
			buffer.append(": ");
			buffer.append(getSourceMACAddress().toString() + "->" +
				getDestinationMACAddress().toString());
			buffer.append(", v="+getCDPVersion());
			buffer.append(", ttl="+getCDPTimeToLive());
			buffer.append(" l="+CDP_HEADER_LEN+","+getCDPData().length);
			buffer.append(']');
			return buffer.toString();
			
		}
	}
}
