// $Id: EthernetPacket.java,v 1.3 2008/10/20 14:43:41 rickwall Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.net;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import net.sourceforge.jpcap.util.AnsiEscapeSequences;
import net.sourceforge.jpcap.util.Timeval;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.zip.CRC32;


/**
 * An Ethernet packet.
 * <p>
 * Contains link-layer header and data payload encapsulated by an Ethernet
 * packet.
 * <p>
 * IP and ARP protocols are supported as subclasses, as are various link-layer and SNAP subclasses.
 *
 * @author Patrick Charles and Jonas Lehmann
 * @author Esmond Pitt
 * @version $Revision: 1.3 $
 * @lastModifiedBy $Author: rickwall $
 * @lastModifiedAt $Date: 2008/10/20 14:43:41 $
 */
public class EthernetPacket extends Packet implements EthernetFields
{
	static final long serialVersionUID = -3211998971645590651L;

  /**
   * The entire packet data.
   * @deprecated Do not use this field: it will be removed in a future release.
   * @see #toByteArray()
   */
  protected byte[] _bytes;
  /**
   * The entire packet data wrapped in a ByteBuffer.
   * @since 2.0.0
   */
  protected transient ByteBuffer	_byteBuffer;

  /**
   * Offset from beginning of byte array where the data payload 
   * (e.g. IP packet) starts. The size of the Ethernet header.
   */
  protected int _ethOffset;

  /**
   * Time the packet was captured off the wire.
   */
  protected Timeval _timeval;

  /**
   * Construct a new ethernet packet.
   * <p>
   * For the purpose of Jpcap, when the type of ethernet packet is 
   * recognized as a protocol for which a class exists network library, 
   * a more specific class like IPPacket or ARPPacket is instantiated.
   *
   * @param lLen link-layer header length
   * @param bytes Packet data
   */
  public EthernetPacket(int lLen, byte [] bytes) {
    this(lLen, bytes, null);
  }

  /**
   * Construct a new ethernet packet, including the capture time.
   *
   * @param lLen Length of link-layer data
   * @param bytes Packet data bytes
   * @param tv Timeval when captured, or null
   *
   */
  public EthernetPacket(int lLen, byte [] bytes, Timeval tv) {
	  this(lLen, ByteBuffer.wrap(bytes), tv);
  }

  /**
   * Construct a new ethernet packet, including the capture time.
   *
   * @param lLen Length of link-layer data
   * @param byteBuffer Packet data in a ByteBuffer
   * @param tv Timeval when captured, or null
   *
   */
  public EthernetPacket(int lLen, ByteBuffer byteBuffer, Timeval tv) {
	  this._ethOffset = this._ethernetHeaderLength = lLen;
	  this._timeval = tv;
	  this._byteBuffer = byteBuffer;
	  if (byteBuffer.hasArray())
		  this._bytes = byteBuffer.array();
	  /*
	   * Removed EJP 18 August 2008. See #toByteArray().
	   * Doing it here is a performance penalty and there is now no code that relies
	   * on _bytes.
	  else
	  {
		  this._bytes = new byte[_byteBuffer.limit()];
		  _byteBuffer.position(0);
		  byteBuffer.get(_bytes);
	  }
	   */
	  _byteBuffer.position(0);
  }
  
  /**
   * {@inheritDoc}
   */
  public final int	getPacketLength()
  {
	  return _byteBuffer.limit();
  }
  
  /**
   * {@inheritDoc}
   */
  public final byte[]	toByteArray()
  {
	  if (_bytes == null)
	  {
		  _bytes = new byte[_byteBuffer.limit()];
		  _byteBuffer.position(0);
		  _byteBuffer.get(_bytes);
	  }
	  return _bytes;
  }
  
  /**
   * {@inheritDoc}
   */
  public final ByteBuffer	getPacketData()
  {
	  return _byteBuffer;
  }

  // Ethernet header length, set in constructor
  private int _ethernetHeaderLength;
  
  /** 
   * Fetch the Ethernet header length in bytes.
   */
  public final int getEthernetHeaderLength() {
    return _ethernetHeaderLength;
  }

  /** 
   * {@inheritDoc}
   */
  public int getHeaderLength() {
    return getEthernetHeaderLength();
  }

  /**
   * Fetch the ethernet header as a byte array.
   * @return the Ethernet header
   */
  public final byte[] getEthernetHeader() {
	  byte[]	header = new byte[getEthernetHeaderLength()];
	  _byteBuffer.position(0);
	  _byteBuffer.get(header);
	  return header;
  }

  /**
   * {@inheritDoc}
   */
  public byte[] getHeader() {
    return getEthernetHeader();
  }

  /**
   * Fetch the Ethernet data as a byte array.
   * @return the Ethernet payload
   */
  public final byte[] getEthernetData() {
	  _byteBuffer.position(getEthernetHeaderLength());
	  byte[]	data = new byte[_byteBuffer.remaining()];
	  _byteBuffer.get(data);
	  return data;
  }

  /**
   * {@inheritDoc}
   */
  public byte[] getData() {
    return getEthernetData();
  }

  /**
   * Fetch the MAC address of the host where the packet originated from.
   * @return the source MAC address, in 00:11:22:33:44:55 format
   * @deprecated Please use {@link #getSourceMACAddress}
   */
  public String getSourceHwAddress() {
	  return getSourceMACAddress().toString();
  }

  /**
   * Fetch the MAC address of the host where the packet originated from.
   * @return the source MAC address.
   * @since 2.0.0
   */
  public MACAddress getSourceMACAddress() {
	  byte[]	address = new byte[MACAddress.WIDTH];
	  _byteBuffer.position(ETH_SRC_POS);
	  _byteBuffer.get(address);
	  return new MACAddress(address);
  }
  
  /**
   * Set the source hardware address.
   * @param macAddress Source MAC address.
   * @since 2.0.0
   */
  public void	setSourceMACAddress(MACAddress macAddress)
  {
	  _byteBuffer.position(ETH_SRC_POS);
	  _byteBuffer.put(macAddress.getAddress());
  }

  /**
   * Fetch the MAC address of the host where the packet originated from.
   * @return the destination MAC address, in 00:11:22:33:44:55 format
   * @deprecated Please use {@link #getDestinationMACAddress}
   */
  public String getDestinationHwAddress() {
	  return getDestinationMACAddress().toString();
  }

  /**
   * Fetch the MAC address of the host where the packet originated from.
   * @return the destination MAC address.
   * @since 2.0.0
   */
  public MACAddress getDestinationMACAddress() {
	  byte[]	address = new byte[MACAddress.WIDTH];
	  _byteBuffer.position(ETH_DST_POS);
	  _byteBuffer.get(address);
	  return new MACAddress(address);
  }
  
  /**
   * Set the destination hardware address.
   * @param macAddress Destination MAC address.
   * @since 2.0.0
   */
  public void	setDestinationMACAddress(MACAddress macAddress)
  {
	  _byteBuffer.position(ETH_DST_POS);
	  _byteBuffer.put(macAddress.getAddress());
  }

  /**
   * Fetch the Ethernet protocol.
   * @return the Ethernet protocol word
   */
  public final int getEthernetProtocol() {
	return _byteBuffer.getShort(ETH_CODE_POS) & 0xffff;
  }
  
  /**
   * Set the Ethernet protocol
   * @param protocol Protocol word
   * @since 2.0.0
   */
  public final void	setEthernetProtocol(int protocol)
  {
	  _byteBuffer.putShort(ETH_CODE_POS, (short)protocol);
  }

  /**
   * {@inheritDoc}
   */
  public int getProtocol() {
    return getEthernetProtocol();
  }

  /**
   * {@inheritDoc}
   */
  public final Timeval getTimeval() {
    return _timeval;
  }

  @Override
  public final void setTimeval(Timeval tv) {
	  _timeval = tv;
  }
  
  /**
   * Return the Ethernet CRC-32.
   * This is recomputed each time, as the actual Ethernet CRC is removed by the Ethernet layer
   * before it gets into Jpcap, and it is recomputed and appended by the Ethernet layer
   * when sending.
   * For this reason no <code>setEthernetCRC32()</code> method is provided,
   * and {@link #setChecksums()} doesn't re-compute it.
   *
   * @return the CRC-32.
   */
  public int	getEthernetCRC32()
  {
	  // Handily, the Ethernet CRC-32 is the one supported by Java.
	  CRC32	crc32 = new CRC32();
	  byte[]	bytes = new byte[_byteBuffer.limit()];
	  _byteBuffer.position(0);
	  _byteBuffer.get(bytes);
	  crc32.update(bytes, 0, bytes.length);
	  return (int)crc32.getValue();
  }
  
  /**
   * {@inheritDoc}
   *<p>
   * After calling this method the {@link #hashCode()} of this object will probably change.
   */
  public void	setChecksums()
  {
	  // Forget cache
	  hashCode = 0;	// forget old hashCode
	  // Nothing else to do here, the Ethernet checksum is takem care of by the network stack.
	  // super.setChecksums();	// is abstract, this is the bottom level
  }
  
  private int	hashCode;
  
  /**
   * {@inheritDoc}
   *<p>
   * This implementation returns the Ethernet CRC-32.
   */
  public int	hashCode()
  {
	  if (hashCode == 0)
		  hashCode = getEthernetCRC32();
	  return hashCode;
  }

  /**
   * {@inheritDoc}
   *<p>
   * This implementation returns true iff the hashCodes, i.e. the CRC-32s, are equal.
   */
  public boolean	equals(Object that)
  {
	  return that instanceof EthernetPacket && this.hashCode() == ((EthernetPacket)that).hashCode();
  }
  
  /**
   * {@inheritDoc}
   */
  public String toString() {
    return toColoredString(false);
  }

  /**
   * {@inheritDoc}
   */
  public String toColoredString(boolean colored) {
    StringBuffer buffer = new StringBuffer();
    buffer.append('[');
    if(colored) buffer.append(getColor());
    buffer.append("EthernetPacket");
    if(colored) buffer.append(AnsiEscapeSequences.RESET);
    buffer.append(": ");
    buffer.append(getSourceHwAddress() + "->" + 
                  getDestinationHwAddress());
    buffer.append(" proto=0x" + Integer.toHexString(getProtocol()));
    buffer.append(" l=" + getEthernetHeaderLength()); // + "," + data.length);
    buffer.append(']');

    return buffer.toString();
  }

  /**
   * {@inheritDoc}
   */
  public String getColor() {
    return AnsiEscapeSequences.DARK_GRAY;
  }
  
  private void	writeObject(ObjectOutputStream out) throws IOException
  {
	 if (_bytes.length != _byteBuffer.limit())
		 _bytes = new byte[_byteBuffer.limit()];
	_byteBuffer.position(0);
	_byteBuffer.get(_bytes);
	out.defaultWriteObject();
  }
  
  /**
   * @serialData This class is serialized as per default serialization;
   * however {@link #readObject} and {@link #writeObject} methods are
   * present to handle the transient ByteBuffer member.
   */
  private Object	readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
  {
	in.defaultReadObject();
	_byteBuffer = ByteBuffer.wrap(_bytes);
	return this;
  }

  private static final String _rcsid = 
    "$Id: EthernetPacket.java,v 1.3 2008/10/20 14:43:41 rickwall Exp $";
}
