// $Id: CaptureTest.java,v 1.1 2008/08/18 05:34:49 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Copyright (C) 2008, Esmond Pitt                                         *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.capture;

import java.io.File;
import java.io.IOException;
import java.net.NetworkInterface;
import java.util.Arrays;
import java.util.Enumeration;
import net.sourceforge.jpcap.net.Packet;
import net.sourceforge.jpcap.net.RawPacket;


/**
 * Class to exercise PacketCapture.
 * <p>
 * Testing of PacketCapture could easily be implemented in the same class's
 * main method, but since the tester is a client and must implement the 
 * listener interface, seems clearer to decouple the two.
 *
 * @author Patrick Charles and Jonas Lehmann
 * @author Esmond Pitt
 * @version $Revision: 1.1 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/18 05:34:49 $
 */
public class CaptureTest implements PacketListener, RawPacketListener
{
  protected static String NAME = "CaptureTest";
  protected static int TEST_COUNT = 2;
	private File	file = new File("test_captures.cap");

  /**
   * Create a new tester.
   */
  public CaptureTest() {
  }


  // implementation of PacketListener interface
  public void	packetArrived(Packet packet)
  {
    System.err.println(NAME + ": " + packet + " arrived");
  }
  
  // implementation of RawPacketListener interface

  public void rawPacketArrived(RawPacket rawPacket) {
    System.err.println(NAME + ": " + rawPacket + " arrived");
  }


  // implementation of test logic

  /**
   * Execute the test.
   */
  public void runTest() {
  
	  System.out.println("Library version: "+PacketCapture.getLibVersion());
	  
	  //-- create PacketCapture system
    System.err.println(NAME + ": instantiating PacketCapture object.. ");
    PacketCapture pc = new PacketCapture();
    System.err.println(NAME + ": PacketCapture instantiated ok");

    //-- set the capture device
    String device = null;
    try {
      System.err.print(NAME + ": detecting capture device.. ");
      device = pc.findDevice();
      System.err.println(device+":network="+Integer.toHexString(pc.getNetwork(device))+":netmask="+Integer.toHexString(pc.getNetmask(device)));
    }
    catch(CaptureDeviceNotFoundException e) {
      // usually occurs if no network devices are available on the system
      System.err.println(e);
      System.exit(1);
    }
    catch(CaptureConfigurationException e) {
      // Shouldn't happen unless the device name is wrong
      System.err.println(e);
      System.exit(1);
    }

    //-- open the capture device
    try {
      System.err.print(NAME + ": opening capture device.. ");
      pc.open(device, true);
      System.err.println(device);
    }
    catch(CaptureDeviceOpenException e) {
      // usually occurs if the user doesn't have sufficient privileges
      System.err.println(e);
      System.exit(1);
    }
	
	try
	{
		// Exercise link layer type API
		System.err.println("Link layer type="+pc.getLinkLayerDescription());
	}
	catch (CaptureConfigurationException exc)
	{
      exc.printStackTrace();
      System.exit(1);
	}

    //-- set filters
    try {
      String filter = ""; // nothing filtered by default
      System.err.println(NAME + ": setting filter to '" + filter + "'.. ");
      pc.setFilter(filter, true);
      System.err.println(NAME + ": filter compiled and activated ok");
    }
    catch(InvalidFilterException e) {
      System.err.println(e);
      System.exit(1);
    }

    //-- register the tester as a packet listener
    System.err.print(NAME + ": registering as a packet listener.. ");
    pc.addPacketListener(this);
    System.err.println("ok");

	//-- register the tester as a raw packet listener
    System.err.print(NAME + ": registering as a raw packet listener.. ");
    pc.addRawPacketListener(this);
    System.err.println("ok");

    //-- capture packets
    try {
      int count = TEST_COUNT;
      System.err.println(NAME + ": waiting for " + count + " packet(s).. ");
      pc.capture(count);
      System.err.println(NAME + ": done capturing.");
    }
    catch(CapturePacketException e) {
		e.printStackTrace();
      System.exit(1);
    }

    //-- dump statistics
    System.err.println(pc.getStatistics());
	
	pc.close();
	
	// Test offline functions
	try
	{
		testLookupDevice();
		testDump();
		testCaptureTo();
		testCaptureToNet();
	}
	catch (Exception exc)
	{
		exc.printStackTrace();
		System.exit(1);
	}
	
  }

  // Exercise same as above but via the findLookupDevice() method,
  // which also exercises listLookupDevices().
  void	testLookupDevice() throws Exception
  {
	  
    PacketCapture pc = null;

    //-- set the capture device
    CaptureDevice device = null;
    try {
      System.err.print(NAME + ": detecting capture device.. ");
      device = PacketCapture.findCaptureDevice();
      System.err.println(device);
    }
    catch(CaptureDeviceNotFoundException e) {
      // usually occurs if no network devices are available on the system
      System.err.println(e);
      System.exit(1);
    }
    catch(CaptureConfigurationException e) {
      // Shouldn't happen unless the device name is wrong
      System.err.println(e);
      System.exit(1);
    }

    //-- open the capture device
    try {
      System.err.print(NAME + ": opening capture device.. ");
      pc = new PacketCapture(device, true);
      System.err.println(device);
    }
    catch(CaptureDeviceOpenException e) {
      // usually occurs if the user doesn't have sufficient privileges
      System.err.println(e);
      System.exit(1);
    }
	
	try
	{
		// Exercise link layer type API
		System.err.println("Link layer type="+pc.getLinkLayerDescription());
	}
	catch (CaptureConfigurationException exc)
	{
      exc.printStackTrace();
      System.exit(1);
	}

    //-- set filters
    try {
      String filter = ""; // nothing filtered by default
      System.err.println(NAME + ": setting filter to '" + filter + "'.. ");
      pc.setFilter(filter, true);
      System.err.println(NAME + ": filter compiled and activated ok");
    }
    catch(InvalidFilterException e) {
      System.err.println(e);
      System.exit(1);
    }

    //-- register the tester as a packet listener
    System.err.print(NAME + ": registering as a packet listener.. ");
    pc.addPacketListener(this);
    System.err.println("ok");

	//-- register the tester as a raw packet listener
    System.err.print(NAME + ": registering as a raw packet listener.. ");
    pc.addRawPacketListener(this);
    System.err.println("ok");

    //-- capture packets
    try {
      int count = TEST_COUNT;
      System.err.println(NAME + ": waiting for " + count + " packet(s).. ");
      pc.capture(count);
      System.err.println(NAME + ": done capturing.");
    }
    catch(CapturePacketException e) {
		e.printStackTrace();
      System.exit(1);
    }

    //-- dump statistics
    System.err.println(pc.getStatistics());	  
	
	pc.close();
  }
  
// Exercise offline functions
  public void	testOffline() throws Exception
  {
	CaptureDevice[]	devices = PacketCapture.lookupCaptureDevices();
	System.out.println("Devices="+Arrays.asList(devices));
	// Open the first device in the list.
	// This API doesn't return 'down' devices, so that should be OK.
	PacketCapture	pc = new PacketCapture(devices[0], true);
	System.out.println("opened "+devices[0]);
	
	// Test output stream
	PacketOutputStream	out = pc.getOutputStream(file);
	System.out.println("Opened dump file "+file);
	
	// Test input stream
	PacketInputStream	in = pc.getInputStream();
	System.out.println("got packet input stream");
	for (int i = 0; i < 10; i++)
	{
		Packet packet = in.readPacket();
		System.out.println("captured packet #"+(i+1)+": "+packet);
		out.writePacket(packet);
		System.out.println("Dumped packet #"+(i+1));
	}
	out.close();
	System.out.println("Closed the dump file");
	pc.close();
	System.out.println("Closed the captor");
  }
  
  // test captureTo() functions to file
  void	testCaptureTo() throws JpcapException, IOException
  {
	PacketCapture	pc = new PacketCapture(file);
	System.out.println("opened captor for dump file "+file);
	
	File	outFile = new File("captureTo_out.cap");
	PacketOutputStream	out = pc.getOutputStream(outFile);
	
	System.out.println("Capturing from "+file+" to "+out);

	pc.captureTo(-1, out);
	
	System.out.println("Capture finished: "+pc.getStatistics());
	
	out.close();
	System.out.println("Closed the dump file");
	
	pc.close();	
	System.out.println("Closed the captor");
	  
  }
  
  /**
   * test captureTo() functions to net.
   * You can really only see if this works if
   * (a) it doesn't crash, and
   * (b) the same traffic as in the file is sniffed by a sniffer.
   */
  void	testCaptureToNet() throws JpcapException, IOException
  {
	PacketCapture	pc = new PacketCapture(file);
	System.out.println("opened captor to read dump file "+file);
	
	CaptureDevice	device = PacketCapture.findCaptureDevice();
	PacketCapture	pc2 = new PacketCapture(device, false);
	System.out.println("opened captor to send to network device "+device);
	
	PacketOutputStream	out = pc2.getOutputStream();
	
	System.out.println("Capturing from "+file+" to "+out);

	pc.captureTo(-1, out);
	
	System.out.println("Capture finished: "+pc.getStatistics());
	
	out.close();
	System.out.println("Closed the dump file");

	pc2.close();
	System.out.println("closed the network output captor");
	pc.close();	
	System.out.println("Closed the savefile input captor");
	  
  }


  /**
   * Simple test to exercise PacketCapture.
   */
  public static void main(String [] args) throws IOException {
	  // Just to prove that java.net.NetworkInterface has nothing to do with CaptureDevice.
	  Enumeration<NetworkInterface>	nwifs = NetworkInterface.getNetworkInterfaces();
	  while (nwifs.hasMoreElements())
		  System.out.println(nwifs.nextElement());
	  
    JWinPcapTest ct = new JWinPcapTest();
    ct.runTest();
  }


  private static final String _rcsid = 
    "$Id: CaptureTest.java,v 1.1 2008/08/18 05:34:49 esmondpitt Exp $";
}
